local constants = {}
constants.SMALL = 1
constants.LARGE = 2

local function prepare_player_data(player)
	local data = {}
	data.notepad_frame = nil
	data.text_box = nil
	data.expand_button = nil
	data.content = ""
	data.location_x = 0
	data.location_y = 0
	data.size = constants.LARGE
	data.has_location = false
	global.players[player.index] = data
end

local function initialize()
	global.players = {}
	for _, player in pairs(game.players) do
		prepare_player_data(player)
	end
end

local function player_created(event)
	prepare_player_data(game.get_player(event.player_index))
end

local function player_removed(event)
	table.remove(global.players, event.player_index)
end

local function get_player_data(player)
	return global.players[player.index]
end

local function update_notepad_size(player)
	local data = get_player_data(player)
	local screen_size = player.display_resolution
	local scale = player.display_scale
	if data.size==constants.SMALL then
		data.expand_button.sprite = "utility/expand"
		data.expand_button.hovered_sprite = "utility/expand_dark"
		data.text_box.style.size = { screen_size.width/9/scale, (screen_size.height/3-24)/scale }
	else
		data.expand_button.sprite = "utility/collapse"
		data.expand_button.hovered_sprite = "utility/collapse_dark"
		data.text_box.style.size = { screen_size.width/3/scale, (screen_size.height/2-24)/scale }
	end
	data.expand_button.clicked_sprite = data.expand_button.hovered_sprite
end

local function create_notepad(player)
	local data = get_player_data(player)

	data.notepad_frame = player.gui.screen.add({type="frame", name="notepad", direction="vertical"})
	data.notepad_frame.style.use_header_filler = true

	local title_bar = data.notepad_frame.add({type="flow", direction="horizontal"})
	title_bar.drag_target = data.notepad_frame

	local title_label = title_bar.add({type="label", caption={"mod-name.notepad"}, style="frame_title", ignored_by_interaction=true})

	local header = title_bar.add({type="empty-widget", style="draggable_space_header", ignored_by_interaction=true})
	header.style.height = 24
	header.style.horizontally_stretchable = true

	data.expand_button = title_bar.add{type="sprite-button", name="notepad-size", sprite="utility/expand", hovered_sprite="utility/expand_dark", clicked_sprite="utility/expand_dark", style="frame_action_button"}

	title_bar.add{type="sprite-button", name="notepad-close", sprite="utility/close_white", hovered_sprite="utility/close_black", clicked_sprite="utility/close_black", style="frame_action_button"}

	data.text_box = data.notepad_frame.add({type="text-box", name="notepad-text"})
	data.text_box.word_wrap = true
	data.text_box.style.horizontally_stretchable = true
	data.text_box.style.vertically_stretchable = true

	data.text_box.text = data.content

	update_notepad_size(player)

	if data.has_location then
		data.notepad_frame.location = { data.location_x, data.location_y }
	else
		data.notepad_frame.force_auto_center()
	end

	data.text_box.focus()
end

local function close_notepad(player)
	local data = get_player_data(player)

	if data.notepad_frame then
		if data.notepad_frame.valid then
			data.notepad_frame.destroy()
		end
		data.notepad_frame = nil
		data.text_box = nil
	end
end

local function toggle_notepad(event)
	local player = game.get_player(event.player_index)
	local data = get_player_data(player)

	if data.notepad_frame then
		close_notepad(player)
		return
	end

	create_notepad(player)
end

local function gui_location_changed(event)
	if event.element.name=="notepad" then
		local player = game.get_player(event.player_index)
		local data = get_player_data(player)
		data.location_x = event.element.location.x
		data.location_y = event.element.location.y
		data.has_location = true
	end
end

local function gui_text_changed(event)
	if event.element.name=="notepad-text" then
		local player = game.get_player(event.player_index)
		get_player_data(player).content = event.text
	end
end

local function gui_clicked(event)
	local player = game.get_player(event.player_index)
	if event.element.name=="notepad-close" then
		close_notepad(player)
	elseif event.element.name=="notepad-size" then
		local data = get_player_data(player)
		if data.size==constants.SMALL then
			data.size = constants.LARGE
		else
			data.size = constants.SMALL
		end
		update_notepad_size(player)
	end
end

script.on_init(initialize)
script.on_event(defines.events.on_player_created, player_created)
script.on_event(defines.events.on_player_removed, player_removed)
script.on_event("toggle-notepad", toggle_notepad)
script.on_event(defines.events.on_gui_location_changed, gui_location_changed)
script.on_event(defines.events.on_gui_text_changed, gui_text_changed)
script.on_event(defines.events.on_gui_click, gui_clicked)
